@@ -22,6 +22,7 @@
 #include <config.h>
 
 #include <glib.h>
+#include <glib/gstdio.h>
 
 #include "terminal-intl.h"
 
@@ -33,7 +34,6 @@
 #include "terminal-window.h"
 #include "terminal-util.h"
 #include "profile-editor.h"
-#include "encoding.h"
 #include <bonobo-activation/bonobo-activation-activate.h>
 #include <bonobo-activation/bonobo-activation-register.h>
 #include <bonobo/bonobo-exception.h>
@@ -82,7 +82,8 @@ typedef struct
   GList   *initial_windows;
   gboolean default_window_menubar_forced;
   gboolean default_window_menubar_state;
-  gboolean default_start_fullscreen;
+  gboolean default_fullscreen;
+  gboolean default_maximize;
   char    *default_role;
   char    *default_geometry;
   char    *default_working_dir;
@@ -119,6 +120,7 @@ typedef struct
   gboolean menubar_state;
 
   gboolean start_fullscreen;
+  gboolean start_maximized;
 
   char *geometry;
   char *role;
@@ -167,6 +169,7 @@ initial_window_new (const char *profile,
   iw->force_menubar_state = FALSE;
   iw->menubar_state = FALSE;
   iw->start_fullscreen = FALSE;
+  iw->start_maximized = FALSE;
   iw->geometry = NULL;
   iw->role = NULL;
   
@@ -204,7 +207,8 @@ apply_defaults (OptionParsingResults *results,
       results->default_window_menubar_forced = FALSE;
     }
 
-  iw->start_fullscreen = results->default_start_fullscreen;
+  iw->start_fullscreen |= results->default_fullscreen;
+  iw->start_maximized |= results->default_maximize;
 }
 
 static InitialWindow*
@@ -283,12 +287,9 @@ option_command_callback (const gchar *option_name,
 {
   OptionParsingResults *results = data;
   InitialTab *it;
-  GError *err;
-  char   **exec_argv;
-  error = NULL;
-  exec_argv = NULL;
+  GError *err = NULL;
+  char  **exec_argv;
 
-  err = NULL;
   if (!g_shell_parse_argv (value, NULL, &exec_argv, &err))
     {
       g_set_error(error,
@@ -520,6 +521,26 @@ option_hide_menubar_callback (const gchar *option_name,
   return TRUE;
 }
 
+static gboolean
+option_maximize_callback (const gchar *option_name,
+                          const gchar *value,
+                          gpointer     data,
+                          GError     **error)
+{
+  OptionParsingResults *results = data;
+  InitialWindow *iw;
+
+  if (results->initial_windows)
+    {
+      iw = g_list_last (results->initial_windows)->data;
+      iw->start_maximized = TRUE;
+    }
+  else
+    results->default_maximize = TRUE;
+
+  return TRUE;
+}
+
 static gboolean
 option_fullscreen_callback (const gchar *option_name,
                             const gchar *value,
@@ -535,7 +556,7 @@ option_fullscreen_callback (const gchar *option_name,
       iw->start_fullscreen = TRUE;
     }
   else
-    results->default_start_fullscreen = TRUE;
+    results->default_fullscreen = TRUE;
 
   return TRUE;
 }
@@ -708,10 +729,11 @@ option_parsing_results_new (int *argc, char **argv)
 
   results = g_slice_new0 (OptionParsingResults);
 
-  results->default_window_menubar_forced = 0;
-  results->default_window_menubar_state = 1;
-  results->default_start_fullscreen = 0;
-  results->execute = 0;
+  results->default_window_menubar_forced = FALSE;
+  results->default_window_menubar_state = TRUE;
+  results->default_fullscreen = FALSE;
+  results->default_maximize = FALSE;
+  results->execute = FALSE;
   results->use_factory = TRUE;
 
   results->startup_id = NULL;
@@ -910,8 +932,9 @@ find_screen_by_display_name (const char *display_name,
         {
           GdkDisplay *disp = l->data;
 
-          /* compare without the screen number part */
-          if (strncmp (gdk_display_get_name (disp), display_name, period - display_name) == 0)
+          /* compare without the screen number part, if present */
+          if ((period && strncmp (gdk_display_get_name (disp), display_name, period - display_name) == 0) ||
+              (period == NULL && strcmp (gdk_display_get_name (disp), display_name) == 0))
             {
               display = disp;
               break;
@@ -983,16 +1006,17 @@ new_terminal_with_options (TerminalApp *app,
       if (results->startup_id)
         terminal_window_set_startup_id (window, results->startup_id);
 
+      /* Overwrite the default, unique window role set in terminal_window_init */
       if (iw->role)
         gtk_window_set_role (GTK_WINDOW (window), iw->role);
-      else
-        terminal_util_set_unique_role (GTK_WINDOW (window), "gnome-terminal");
 
       if (iw->force_menubar_state)
         terminal_window_set_menubar_visible (window, iw->menubar_state);
 
       if (iw->start_fullscreen)
         gtk_window_fullscreen (GTK_WINDOW (window));
+      if (iw->start_maximized)
+        gtk_window_maximize (GTK_WINDOW (window));
 
       /* Now add the tabs */
       for (lt = iw->tabs; lt != NULL; lt = lt->next)
@@ -1111,6 +1135,9 @@ slowly_and_stupidly_obtain_timestamp (Display *xdisplay)
   return event.xproperty.time;
 }
 
+/* Evil hack alert: this is exported from libgconf-2 but not in a public header */
+extern gboolean gconf_ping_daemon (void);
+         
 int
 main (int argc, char **argv)
 {
@@ -1120,6 +1147,7 @@ main (int argc, char **argv)
   char **argv_copy;
   const char *startup_id;
   const char *display_name;
+  const char *home_dir;
   GdkDisplay *display;
   GnomeProgram *program;
   OptionParsingResults *parsing_results;
@@ -1128,6 +1156,8 @@ main (int argc, char **argv)
   bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
   textdomain (GETTEXT_PACKAGE);
 
+  g_thread_init (NULL);
+
   argc_copy = argc;
   /* we leave empty slots, for --startup-id, --display and --default-working-directory */
   argv_copy = g_new0 (char *, argc_copy + 7);
@@ -1137,10 +1167,10 @@ main (int argc, char **argv)
 
   parsing_results = option_parsing_results_new (&argc, argv);
   startup_id = g_getenv ("DESKTOP_STARTUP_ID");
-  if (startup_id != NULL && *startup_id != '\0')
+  if (startup_id != NULL && startup_id[0] != '\0')
     {
       parsing_results->startup_id = g_strdup (startup_id);
-      putenv ((char *) "DESKTOP_STARTUP_ID=");
+      g_unsetenv ("DESKTOP_STARTUP_ID");
     }
   
   gtk_window_set_auto_startup_notification (FALSE); /* we'll do it ourselves due
@@ -1212,6 +1242,15 @@ main (int argc, char **argv)
   g_strfreev (argv_copy);
   argv_copy = NULL;
 
+  /* If the gconf daemon isn't available (e.g. because there's no dbus
+   * session bus running), we'd crash later on. Tell the user about it
+   * now, and exit. See bug #561663. */
+  if (!gconf_ping_daemon ())
+    {
+      g_printerr ("Failed to contact the GConf daemon; exiting.\n");
+      exit (1);
+    }
+
   gtk_window_set_default_icon_name (GNOME_TERMINAL_ICON_NAME);
  
   g_assert (parsing_results->post_execute_args == NULL);
@@ -1225,7 +1264,14 @@ main (int argc, char **argv)
 
   initialization_complete = TRUE;
   handle_new_terminal_events ();
-  
+
+  /* Now change directory to $HOME so we don't prevent unmounting, e.g. if the
+   * factory is started by nautilus-open-terminal. See bug #565328.
+   */
+  home_dir = g_get_home_dir ();
+  if (home_dir)
+    g_chdir (home_dir);
+
   gtk_main ();
 
   terminal_app_shutdown ();
@@ -1354,6 +1400,15 @@ get_goption_context (OptionParsingResults *parsing_results)
       N_("Turn off the menubar for the last-specified window; applies to only one window; can be specified once for each window you create from the command line."),
       NULL
     },
+    {
+      "maximize",
+      0,
+      G_OPTION_FLAG_NO_ARG,
+      G_OPTION_ARG_CALLBACK,
+      option_maximize_callback,
+      N_("Set the last-specified window into maximized mode; applies to only one window; can be specified once for each window you create from the command line."),
+      NULL
+    },
     {
       "full-screen",
       0,
